from typing import (
    Optional,
)

from edu_rdm_integration.adapters.strategies import (
    WebEduSyncBaseRunnerLazySavingPredefinedQueueFunctionImplementationStrategy,
)
from edu_rdm_integration.collect_data.non_calculated.base.caches import (
    BaseCollectingExportedDataFunctionCacheStorage,
    BaseCollectingExportedDataRunnerCacheStorage,
)
from edu_rdm_integration.collect_data.non_calculated.base.errors import (
    BaseCollectingExportedDataError,
)
from edu_rdm_integration.collect_data.non_calculated.base.functions import (
    BaseCollectingExportedDataFunction,
)
from edu_rdm_integration.collect_data.non_calculated.base.helpers import (
    BaseCollectingExportedDataFunctionHelper,
    BaseCollectingExportedDataRunnerHelper,
)
from edu_rdm_integration.collect_data.non_calculated.base.managers import (
    BaseCollectingExportedDataRunnerManager,
)
from edu_rdm_integration.collect_data.non_calculated.base.presenters import (
    BaseCollectingExportedDataResultPresenter,
)
from edu_rdm_integration.collect_data.non_calculated.base.results import (
    BaseCollectingExportedDataFunctionResult,
    BaseCollectingExportedDataRunnerResult,
)
from edu_rdm_integration.collect_data.non_calculated.base.runners import (
    BaseCollectingExportedDataRunner,
)
from edu_rdm_integration.collect_data.non_calculated.base.validators import (
    BaseCollectingExportedDataFunctionValidator,
    BaseCollectingExportedDataRunnerValidator,
)
from edu_rdm_integration.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
)


class CollectingExportedDataFunctionImplementationStrategy(
    WebEduSyncBaseRunnerLazySavingPredefinedQueueFunctionImplementationStrategy,
):
    """
    Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение.

    Используется для функций сбора данных для выгрузки в "Региональная витрина данных".
    """

    @classmethod
    def _prepare_uuid(cls):
        """
        Получение UUID класса. Используется при регистрации сущности в базе данных.

        Если ничего не возвращает, то регистрация в БД не будет произведена.
        """
        return '475eb094-b7fc-422a-a8d2-6cf3e0859cae'

    @classmethod
    def _prepare_verbose_name(cls) -> str:
        """
        Полное наименование для дальнейшей регистрации и отображения пользователю.
        """
        return (
            'Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение '
            'функций сбора данных для выгрузки в "Региональная витрина данных"'
        )

    @classmethod
    def _prepare_tags(cls) -> list[str]:
        """
        Список тегов, по которым сущность можно будет осуществлять поиск.
        """
        return [REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA]

    @classmethod
    def _prepare_key(cls) -> str:
        """
        Возвращает уникальный идентификатор стратегии создания функции.
        """
        return 'REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA'

    @classmethod
    def _prepare_title(cls) -> str:
        """
        Возвращает название стратегии создания функции.
        """
        return (
            'Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение '
            'функций сбора данных для выгрузки в "Региональная витрина данных". Сохранение производится после удачной '
            'работы функции'
        )

    @classmethod
    def _prepare_function_template_name(cls) -> Optional[str]:
        """
        Формирование названия шаблона создания функции.
        """
        return 'function_collect_data_template'

    def _prepare_manager_class(self):
        """
        Устанавливает класс менеджера.
        """
        self._manager_class = BaseCollectingExportedDataRunnerManager

    def _prepare_runner_class(self):
        """
        Устанавливает класс ранера.
        """
        self._runner_class = BaseCollectingExportedDataRunner

    def _prepare_function_class(self):
        """
        Устанавливает класс Функции.
        """
        self._function_class = BaseCollectingExportedDataFunction

    def _prepare_runner_helper_class(self):
        """
        Устанавливает класс помощника ранера.
        """
        self._runner_helper_class = BaseCollectingExportedDataRunnerHelper

    def _prepare_function_helper_class(self):
        """
        Устанавливает класс помощника функции.
        """
        self._function_helper_class = BaseCollectingExportedDataFunctionHelper

    def _prepare_runner_validator_class(self):
        """
        Устанавливает класс валидатора ранера.
        """
        self._runner_validator_class = BaseCollectingExportedDataRunnerValidator

    def _prepare_function_validator_class(self):
        """
        Устанавливает класс валидатора Функции.
        """
        self._function_validator_class = BaseCollectingExportedDataFunctionValidator

    def _prepare_runner_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей ранера.
        """
        self._runner_cache_storage_class = BaseCollectingExportedDataRunnerCacheStorage

    def _prepare_function_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей Функции.
        """
        self._function_cache_storage_class = BaseCollectingExportedDataFunctionCacheStorage

    def _prepare_error_class(self):
        """
        Устанавливает класс ошибки.
        """
        self._error_class = BaseCollectingExportedDataError

    def _prepare_runner_result_class(self):
        """
        Устанавливает класс результата.
        """
        self._runner_result_class = BaseCollectingExportedDataRunnerResult

    def _prepare_function_result_class(self):
        """
        Устанавливает класс результата.
        """
        self._function_result_class = BaseCollectingExportedDataFunctionResult

    def _prepare_result_presenter_class(self):
        """
        Устанавливает класс презентера результата.
        """
        self._result_presenter_class = BaseCollectingExportedDataResultPresenter



"""
Команда для проверки состояния отправленных в витрину данных.

Необязательные параметры:
--period_started_at - дата и время начала периода загрузки данных в витрину
--period_ended_at - дата и время конца периода загрузки данных в витрину
--thread_count - количество потоков для обработки

Пример использования:
python manage.py check_upload_status --period_started_at="01.11.2023 10:00:00" --thread_count=4
"""
from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    TYPE_CHECKING,
    Any,
)
from django.core.cache import (
    cache,
)
from django.core.management.base import (
    BaseCommand,
)

from edu_rdm_integration.consts import (
    DATETIME_FORMAT,
)
from edu_rdm_integration.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.helpers import (
    UploadStatusHelper,
)
from edu_rdm_integration.models import (
    ExportingDataSubStageUploaderClientLog,
)


if TYPE_CHECKING:
    from django.core.management.base import (
        CommandParser,
    )


class Command(BaseCommand):
    """Команда для проверки состояния отправленных в витрину данных."""

    help = 'Команда для проверки состояния отправленных в витрину данных'  # noqa: A003

    def add_arguments(self, parser: 'CommandParser') -> None:
        """Добавляет аргументы парсера."""
        parser.add_argument(
            '--period_started_at',
            action='store',
            dest='period_started_at',
            type=lambda started_at: datetime.strptime(started_at, DATETIME_FORMAT),
            default=datetime.combine(date.today(), time.min),
            help=(
                'Дата и время начала периода загрузки данных в витрину. Значение предоставляется в формате '
                '"дд.мм.гггг чч:мм:сс". По умолчанию, сегодняшний день, время 00:00:00.'
            ),
        )

        parser.add_argument(
            '--period_ended_at',
            action='store',
            dest='period_ended_at',
            type=lambda ended_at: datetime.strptime(ended_at, DATETIME_FORMAT),
            default=datetime.combine(date.today(), time.max),
            help=(
                'Дата и время конца периода загрузки данных в витрину. Значение предоставляется в формате '
                '"дд.мм.гггг чч:мм:сс". По умолчанию, сегодняшний день, время 23:59:59.'
            ),
        )

        parser.add_argument(
            '--thread_count',
            default=1,
            type=int,
            help='Количество потоков для обработки. По умолчанию, 1.',
        )

    def handle(self, *args: tuple[Any], **kwargs: dict[str, Any]) -> None:
        """Обработчик команды."""
        thread_count = kwargs['thread_count']
        if thread_count < 1:
            raise ValueError(f'Количество потоков {thread_count} должно быть больше 0.')

        in_progress_attachment_uploads = ExportingDataSubStageUploaderClientLog.objects.filter(
            created__gte=kwargs['period_started_at'],
            created__lte=kwargs['period_ended_at'],
            is_emulation=False,
            file_upload_status=FileUploadStatusEnum.IN_PROGRESS,
        )

        UploadStatusHelper(in_progress_attachment_uploads, cache).run(thread_count=thread_count)



"""
Команда для проставления расчетов файла . По умолчанию собирается за месяц. Обновление работает асинхронно.

Выполнение:
django-admin async_fix_attachment_size --chunk_size=50 --count_store_size --connection_pool_size=5
 --async_workers_count=10 --date_begin='01-01-2023'

Параметры:
 --date_begin: дата начала сбора данных по сгенерированным файлам в формате ДД-ММ-ГГГГ (по умолчанию месяц назад).
 --date_end: дата окончания сбора данных по сгенерированным файлам в формате ДД-ММ-ГГГГ (по умолчанию текущая дата).
 --chunk_size: количество файлов в одной итерации, по умолчанию 500
 --count_store_size: при указании этого флага, будет произведен расчет размера файла (
 запрос через оs и вычисление размера) - не рекомендуется при большом количестве данных и на дампах с прода
--connection_pool_size: количество одновременных коннектов к базе по умолчанию 10
--async_workers_count количество одновременных асинхронных воркеров-исполнителей по умолчанию 20
"""


import asyncio
from datetime import (
    date,
    datetime,
    timedelta,
)

import asyncpg
from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
    CommandError,
)

from educommon import (
    logger,
)
from educommon.utils.seqtools import (
    make_chunks,
)

from edu_rdm_integration.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.models import (
    ExportingDataSubStageAttachment,
)


DEFAULT_QUERYSET_CHUNK_SIZE = 500
DEFAULT_FILE_SIZE = 10_485_760
DEFAULT_ASYNC_WORKERS_COUNT = 20
DEFAULT_POOL_SIZE = 10
DB_SETTINGS = settings.DATABASES['default']


class Command(BaseCommand):
    """Команда для проставления размера по умолчанию (в байтах) для отправленных файлов."""

    help = (  # noqa: A003
        'Команда для проставления размера по умолчанию (в байтах) для отправленных файлов.'
    )

    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.result = {
            'initial_count_rows_to_fix': 0,
            'errors': [],
            'fixed_rows_count': 0,
        }

    def add_arguments(self, parser):
        """Добавление аргументов запуска команды."""
        parser.add_argument(
            '--chunk_size',
            type=int,
            default=DEFAULT_QUERYSET_CHUNK_SIZE,
            help='Кол-во единовременно изменяемых элементов в запросе'
        )
        parser.add_argument(
            '--count_store_size',
            action='store_true',
            default=False,
            help='Рассчитать размер файлов'
        ),
        parser.add_argument(
            '--connection_pool_size',
            type=int,
            default=DEFAULT_POOL_SIZE,
            help='Кол-во одновременных коннектов к базе'
        )
        parser.add_argument(
            '--async_workers_count',
            type=int,
            default=DEFAULT_ASYNC_WORKERS_COUNT,
            help='Кол-во одновременных асинхронных воркеров-исполнителей'
        )
        parser.add_argument(
            '--date_begin',
            type=lambda d: datetime.strptime(d, '%d-%m-%Y').date(),
            help='Дата начала поиска сгенерированных файлов в формате %%d-%%m-%%Y.',
        )
        parser.add_argument(
            '--date_end',
            type=lambda d: datetime.strptime(d, '%d-%m-%Y').date(),
            help='Дата окончания поиска сгенерированных файлов в формате %%d-%%m-%%Y.',
        )

    def get_attachments_fix_query(
        self,
        chunk_size: int,
        date_begin: date,
        date_end: date,
        count_store_size: bool
    ) -> list[tuple[int, int]]:
        """Формирование id файлов, которым надо обновить размер."""
        logger.info('Определение файлов, которым надо обновить размер.')
        attachment_query = ExportingDataSubStageAttachment.objects.filter(
            exportingdatasubstageuploaderclientlog__file_upload_status=FileUploadStatusEnum.IN_PROGRESS,
            exportingdatasubstageuploaderclientlog__is_emulation=False,
            attachment_size__isnull=True,
            modified__gte=date_begin,
            modified__lte=date_end,
        )
        update_attachments = []

        for chunk in make_chunks(attachment_query.iterator(), chunk_size):
            if count_store_size:
                update_attachments.append(
                   [
                       (
                           attachment.id, attachment.attachment.size
                           if attachment.attachment.field.storage.exists(attachment.attachment.name)
                           else DEFAULT_FILE_SIZE
                       )
                       for attachment in chunk
                   ]
                )
            else:
                update_attachments.append([(attachment.id, DEFAULT_FILE_SIZE) for attachment in chunk])

        return update_attachments

    async def execute_update_sql(
        self,
        pool: asyncpg.Pool,
        data: list[tuple[int, int]]
    ) -> None:
        """Метод для исполнения сгенерированного UPDATE sql запроса."""
        self.result['initial_count_rows_to_fix'] += len(data)
        logger.info('Обновление размера файлов')
        async with pool.acquire() as connection:
            try:
                await connection.executemany(
                    """
                    UPDATE rdm_exporting_data_sub_stage_attachment
                    set attachment_size = $2
                    where id = $1;
                    """, data,
                )
                self.result['fixed_rows_count'] += len(data)
            except Exception as ex:
                self.result['errors'].append(str(ex))

    async def fix_attachment_size(
        self,
        fix_queries: list,
        pool_size: int = None,
        workers_count: int = None
    ) -> None:
        """Асинхронное обновление размера файлов."""
        pool = await asyncpg.create_pool(
            max_size=pool_size,
            min_size=pool_size,
            host=DB_SETTINGS['HOST'],
            port=DB_SETTINGS['PORT'],
            user=DB_SETTINGS['USER'],
            password=DB_SETTINGS['PASSWORD'],
            database=DB_SETTINGS['NAME']
        )
        if len(fix_queries) > workers_count:
            for chunk in make_chunks(fix_queries, workers_count):
                tasks = [self.execute_update_sql(pool, query) for query in chunk]
                await asyncio.gather(*tasks)
        else:
            tasks = [self.execute_update_sql(pool, query) for query in fix_queries]
            await asyncio.gather(*tasks)

    def update_attachment_size(
        self,
        connection_pool_size: int,
        chunk_size: int,
        date_begin: date,
        date_end:date,
        async_workers_count: int,
        count_store_size: bool
    ):
        """Обновление размеров файла."""
        fix_queries = self.get_attachments_fix_query(chunk_size, date_begin, date_end, count_store_size)
        if fix_queries:
            event_loop = asyncio.get_event_loop()
            try:
                event_loop.run_until_complete(
                    self.fix_attachment_size(fix_queries, connection_pool_size, async_workers_count)
                )
            finally:
                event_loop.close()

    def handle(self, *args, **options):
        """Выполнение команды."""
        time_start = datetime.now()

        chunk_size = options['chunk_size']
        connection_pool_size = options['connection_pool_size']
        async_workers_count = options['async_workers_count']
        count_store_size = options['count_store_size']

        date_begin = options.get('date_begin')
        date_end = options.get('date_end') or date.today()

        if not date_begin:
            date_begin = date_end - timedelta(days=30)

        elif date_begin > date_end:
            raise CommandError('Дата начала сборки больше, чем дата окончания')

        logger.info('Обновление размеров файла\n')

        self.update_attachment_size(
            connection_pool_size,
            chunk_size,
            date_begin,
            date_end,
            async_workers_count,
            count_store_size
        )

        delta = datetime.now() - time_start

        logger.info(f'Время выполнения: {delta}\n')

        logger.info(
            f'Начальное кол-во записей для исправления: {self.result["initial_count_rows_to_fix"]}\n'
            f'Исправлено: {self.result["fixed_rows_count"]}\n'
        )

        errors = self.result['errors']
        for error in errors:
            logger.error(error)



# Generated by Django 3.1.14 on 2024-05-22 16:19

from django.db import (
    migrations,
)


def add_exporting_data_sub_stage_status(apps, schema_editor):
    """Добавление нового статуса этапу выгрузки данных."""
    ExportingDataSubStageStatus = apps.get_model('edu_rdm_integration', 'ExportingDataSubStageStatus')  # noqa: N806
    ExportingDataSubStageStatus.objects.get_or_create(
        key='READY_FOR_EXPORT',
        title='Готов к выгрузке'
    )


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration', '0008_transferredentity'),
    ]

    operations = [
        migrations.RunPython(add_exporting_data_sub_stage_status, reverse_code=migrations.RunPython.noop),
    ]



from edu_rdm_integration.adapters.results import (
    WebEduFunctionResult,
    WebEduRunnerResult,
)


class BaseExportDataRunnerResult(WebEduRunnerResult):
    """
    Базовый класс результатов работы ранеров функций выгрузки данных для интеграции с "Региональная витрина данных".
    """


class BaseExportDataFunctionResult(WebEduFunctionResult):
    """
    Базовый класс результатов работы функций выгрузки данных для интеграции с "Региональная витрина данных".
    """



from abc import (
    ABCMeta,
)

from function_tools.presenters import (
    ResultPresenter,
)


class WebEduResultPresenter(ResultPresenter, metaclass=ABCMeta):
    """
    Базовый класс презентеров результатов работы функций продуктов Образования.
    """



# Ссылка на каталог с файлами для загрузки
UPLOADS = 'uploads'

# Префикс идентификаторов записей сущностей специфический для продукта. Должен быть заполнен в продуктах в settings.py
# или в конфигурационном файле
RDM_EXPORT_ENTITY_ID_PREFIX = ''

# Количество записей моделей ЭШ обрабатываемых за одну итерацию сбора данных
RDM_COLLECT_CHUNK_SIZE = 500

# Количество записей моделей обрабатываемых за одну итерацию экспорта
RDM_EXPORT_CHUNK_SIZE = 500

# Количество не экспортированных записей моделей обрабатываемых за одну итерацию обновления поля modified
RDM_UPDATE_NON_EXPORTED_CHUNK_SIZE = 5_000

# Отображение пункта меню "Региональная витрина данных"
RDM_MENU_ITEM = True


# Загрузка данных в Региональную витрину данных (РВД)
# Адрес витрины (schema://host:port)
RDM_UPLOADER_CLIENT_URL = 'http://localhost:8090'

# Мнемоника Витрины
RDM_UPLOADER_CLIENT_DATAMART_NAME = 'test'

# Количество повторных попыток запроса
RDM_UPLOADER_CLIENT_REQUEST_RETRIES = 10

# Таймаут запроса, сек
RDM_UPLOADER_CLIENT_REQUEST_TIMEOUT = 10

# Включить эмуляцию отправки запросов
RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION = True


# Настройка запуска периодической задачи выгрузки данных:
RDM_TRANSFER_TASK_MINUTE = '0'
RDM_TRANSFER_TASK_HOUR = '*/4'
RDM_TRANSFER_TASK_DAY_OF_WEEK = '*'
RDM_TRANSFER_TASK_LOCK_EXPIRE_SECONDS = 60 * 60 * 6

# Настройка запуска периодической задачи статуса загрузки данных в витрину:
RDM_UPLOAD_STATUS_TASK_MINUTE = '*/30'
RDM_UPLOAD_STATUS_TASK_HOUR = '*'
RDM_UPLOAD_STATUS_TASK_DAY_OF_WEEK = '*'
RDM_UPLOAD_STATUS_TASK_LOCK_EXPIRE_SECONDS = 60 * 60 * 2

# Настройка очереди Redis для формирования файлов РВД
RDM_UPLOAD_QUEUE_MAX_SIZE = 500_000_000
# Таймаут для сохранения параметров в общем кеш
RDM_REDIS_CACHE_TIMEOUT_SECONDS = 60 * 60 * 2


# Настройка запуска периодической задачи отправки csv-файлов в витрину:
RDM_UPLOAD_DATA_TASK_MINUTE = '0'
RDM_UPLOAD_DATA_TASK_HOUR = '*/2'
RDM_UPLOAD_DATA_TASK_DAY_OF_WEEK = '*'
RDM_UPLOAD_DATA_TASK_LOCK_EXPIRE_SECONDS = 60 * 60 * 2
# Количество подэтапов для обработки в периодической задаче отправки данных
RDM_UPLOAD_DATA_TASK_EXPORT_STAGES = 500

RDM_RESPONSE_FILE_STATUS = 'success'
